#pragma once

#include <cstdint>

#include "hephaistos/config.hpp"

namespace hephaistos {

/**
 * @brief Major version of hephaistos
*/
HEPHAISTOS_API extern const uint32_t VERSION_MAJOR;
/**
 * @brief Minor version of hephaistos
*/
HEPHAISTOS_API extern const uint32_t VERSION_MINOR;
/**
 * @brief Patch version of hephaistos
*/
HEPHAISTOS_API extern const uint32_t VERSION_PATCH;

}
